
SuperStrict

Framework GLFW.GLFWWindow

Import GLFW.GLFW
Import GLFW.GLFWMonitor
Import GLFW.GLFWOpenGL
Import GLFW.GLFWSystem

Import BRL.StandardIO
Import BRL.LinkedList

Import BRL.Matrix
Import BRL.Quaternion
Import BRL.Vector

' CAMERA STUFF

Local view_matrix:SMat4F
Local projection_matrix:SMat4F

' -----------------------------------------------------------------------------
Const HACK:Int = False				' H A C K    M E !
' -----------------------------------------------------------------------------

Local near:Float
Local far:Float
Local fov:Float
Local position:SVec3F
Local tri_scale:Float

If Not HACK						' Settings used in tutorials
	near		= 0.1
	far			= 10.0
	fov			= Radians (50.0)
	position	= New SVec3F (3.0, 3.0, 3.0)
	tri_scale	= 1.0
Else							' Settings that render something!
	near		= 0.1
	far			= 250.0
	fov			= Radians (50.0)
	position	= New SVec3F (3.5, 3.5, 3.5)
	tri_scale	= 0.5
EndIf

' https://learnopengl.com/Getting-started/Transformations
' https://www.tomdalling.com/blog/modern-opengl/03-matrices-depth-buffering-animation/
' https://github.com/tomdalling/opengl-series/blob/master/source/03_matrices/source/main.cpp

Local TMP_BasicCube:Float [] = [..
	..
	..
	..		'	x		y			z
			-tri_scale, tri_scale, tri_scale,		1.0, 0.0, 0.0, .. ' Near - Red
			tri_scale, -tri_scale, tri_scale,		1.0, 0.0, 0.0, ..
			-tri_scale, -tri_scale, tri_scale,		1.0, 0.0, 0.0, ..
			
			-tri_scale, tri_scale, tri_scale,		0.5, 0.0, 0.0, ..
			tri_scale, tri_scale, tri_scale,		0.5, 0.0, 0.0, ..
			tri_scale, -tri_scale, tri_scale,		0.5, 0.0, 0.0, ..
			
			tri_scale, tri_scale, -tri_scale,		0.0, 1.0, 0.0, .. ' Far - Green
			-tri_scale, -tri_scale, -tri_scale,		0.0, 1.0, 0.0, ..
			tri_scale, -tri_scale, -tri_scale,		0.0, 1.0, 0.0, ..
			
			tri_scale, tri_scale, -tri_scale,		0.0, 0.5, 0.0, ..
			-tri_scale, tri_scale, -tri_scale,		0.0, 0.5, 0.0, ..
			-tri_scale, -tri_scale, -tri_scale,		0.0, 0.5, 0.0, ..

			-tri_scale, tri_scale, -tri_scale,		0.0, 0.0, 1.0, .. ' Left - Blue
			-tri_scale, -tri_scale, tri_scale,		0.0, 0.0, 1.0, ..
			-tri_scale, -tri_scale, -tri_scale,		0.0, 0.0, 1.0, ..
'			
			-tri_scale, tri_scale, -tri_scale,		0.0, 0.0, 0.5, ..
			-tri_scale, tri_scale, tri_scale,		0.0, 0.0, 0.5, ..
			-tri_scale, -tri_scale, tri_scale,		0.0, 0.0, 0.5, ..
			
			tri_scale, tri_scale, tri_scale,		1.0, 1.0, 0.0, .. ' Right - Yellow
			tri_scale, -tri_scale, -tri_scale,		1.0, 1.0, 0.0, ..
			tri_scale, -tri_scale, tri_scale,		1.0, 1.0, 0.0, ..
'			
			tri_scale, tri_scale, tri_scale,		0.5, 0.5, 0.0, ..
			tri_scale, tri_scale, -tri_scale,		0.5, 0.5, 0.0, ..
			tri_scale, -tri_scale, -tri_scale,		0.5, 0.5, 0.0, ..
			
			-tri_scale, tri_scale, -tri_scale,		0.0, 1.0, 1.0, .. ' Top - Cyan
			tri_scale, tri_scale, tri_scale,		0.0, 1.0, 1.0, ..
			-tri_scale, tri_scale, tri_scale,		0.0, 1.0, 1.0, ..
			
			-tri_scale, tri_scale, -tri_scale,		0.0, 0.5, 0.5, ..
			tri_scale, tri_scale, -tri_scale,		0.0, 0.5, 0.5, ..
			tri_scale, tri_scale, tri_scale,		0.0, 0.5, 0.5, ..

			-tri_scale, -tri_scale, tri_scale,		1.0, 0.0, 1.0, .. ' Bottom - Magenta
			tri_scale, -tri_scale, -tri_scale,		1.0, 0.0, 1.0, ..
			-tri_scale, -tri_scale, -tri_scale,		1.0, 0.0, 1.0, ..
			
			-tri_scale, -tri_scale, tri_scale,		0.5, 0.0, 0.5, ..
			tri_scale, -tri_scale, tri_scale,		0.5, 0.0, 0.5, ..
			tri_scale, -tri_scale, -tri_scale,		0.5, 0.0, 0.5]

' Demo window control...

Global AppWindow:TGLFWWindow							' Application window

Global _RetroGraphicsWidth:Int = 800							' Display width
Global _RetroGraphicsHeight:Int = 600							' Display height

'Retro3D (UInt (_RetroGraphicsWidth), UInt (_RetroGraphicsHeight))

'TGLFWWindow.Hint (GLFW_SAMPLES, msaa_level)
TGLFWWindow.Hint (GLFW_CONTEXT_VERSION_MAJOR, 3)
TGLFWWindow.Hint (GLFW_CONTEXT_VERSION_MINOR, 3)
TGLFWWindow.Hint (GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE)

?MacOS ' Ewww...
	TGLFWWindow.Hint (GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE)
?

Local monitor:TGLFWMonitor = Null

'If full_screen Then monitor = TGLFWMonitor.GetPrimaryMonitor ()

AppWindow = New RetroGameWindow.Create (_RetroGraphicsWidth, _RetroGraphicsHeight, "Demo", monitor)

If Not AppWindow
	Print "Failed to create GLFW window!"
	End
EndIf

AppWindow.MakeContextCurrent ()

gladLoadGL (glfwGetProcAddress)

' DEPTH TESTING

glEnable (GL_DEPTH_TEST)
glDepthFunc (GL_LESS)

AppWindow.GetFrameBufferSize (_RetroGraphicsWidth, _RetroGraphicsHeight)

' SHADER PROGRAM

Local program:Int

' MESH

Local vertex_data:Float []				' Mesh vertices (3D points)
Local vao:UInt, vbo:UInt				' VAO and VBO
Local triangle_count:Int				' Pre-calc number of triangles
Local vertex_count:Int					' Pre-calc number of vertices

' ASSIGN CUBE DATA TO MESH

vertex_data = TMP_BasicCube

triangle_count	= (vertex_data.length / ATTRIBUTES_PER_VERTEX) / FLOATS_PER_ATTRIBUTE / VERTICES_PER_TRIANGLE
vertex_count	= triangle_count * VERTICES_PER_TRIANGLE

Const FLOATS_PER_ATTRIBUTE:Int	= 3	' R, G, B and X, Y, Z each contain 3 floats... for now.
Const ATTRIBUTES_PER_VERTEX:Int	= 2	' Vertex data contains both RGB and XYZ attributes
Const VERTICES_PER_TRIANGLE:Int	= 3 ' Just to avoid Magic Number in code!

Global RetroFloatSize:Int		= SizeOf (0:Float) ' Avoiding reallocating a 0 for obtaining pointer each time!

Global VA_Offset:Int			= FLOATS_PER_ATTRIBUTE * RetroFloatSize
Global VA_Stride:Int			= VA_Offset * ATTRIBUTES_PER_VERTEX

' LOAD SHADERS

Local vertex_source_file:String = "vertex.glsl"
Local fragment_source_file:String = "fragment.glsl"

If Not FileType (vertex_source_file)	Then vertex_source_file		= "shaders/" + vertex_source_file
If Not FileType (fragment_source_file)	Then fragment_source_file	= "shaders/" + fragment_source_file

Local vertex_source:String		= LoadString (vertex_source_file)
Local fragment_source:String	= LoadString (fragment_source_file)

' ---------------------------------------------------------------------
' Create vertex shader object...
' ---------------------------------------------------------------------

Local vertex_shader:Int = glCreateShader (GL_VERTEX_SHADER)

' ---------------------------------------------------------------------
' Build and check vertex shader...
' ---------------------------------------------------------------------

glShaderSource	(vertex_shader, 1, vertex_source)
glCompileShader	(vertex_shader)

Local success:Int

glGetShaderiv (vertex_shader, GL_COMPILE_STATUS, Varptr success)

If Not success
	'ShaderCompilationError (fragment_shader, fragment_source, vertex_source_file)
	Print "Failed to compile shader"
	End
EndIf

' ---------------------------------------------------------------------
' Create vertex shader object...
' ---------------------------------------------------------------------

Local fragment_shader:Int = glCreateShader (GL_FRAGMENT_SHADER)

' ---------------------------------------------------------------------
' Build and check fragment shader...
' ---------------------------------------------------------------------

glShaderSource	(fragment_shader, 1, fragment_source)
glCompileShader	(fragment_shader)

glGetShaderiv	(fragment_shader, GL_COMPILE_STATUS, Varptr success)

If Not success
	'ShaderCompilationError (fragment_shader, fragment_source, vertex_source_file)
	Print "Failed to compile shader"
	End
EndIf

' ---------------------------------------------------------------------
' Create shader program and attach compiled shader objects...
' ---------------------------------------------------------------------

program:Int = glCreateProgram ()

glAttachShader	(program, vertex_shader)
glAttachShader	(program, fragment_shader)

' ---------------------------------------------------------------------
' Link and check shader program...
' ---------------------------------------------------------------------

glLinkProgram	(program)

glGetProgramiv	(program, GL_LINK_STATUS, Varptr success)

If Not success
	Print ""
	Print "Shader program linking failed: " + glGetProgramInfoLog (program)
	End
EndIf

' ---------------------------------------------------------------------
' Detach and delete shaders as no longer needed, now part of program...
' ---------------------------------------------------------------------

glDetachShader (program, vertex_shader)
glDetachShader (program, fragment_shader)

glDeleteShader (vertex_shader)
glDeleteShader (fragment_shader)

' ---------------------------------------------------------------------
' Needs to be active before uniforms can be used!
' ---------------------------------------------------------------------

glUseProgram (program)
		
' ---------------------------------------------------------------------
' VAO: Container for raw data...
' ---------------------------------------------------------------------

' Create and bind as current VAO...

glGenVertexArrays (1, Varptr vao)	' NB. 1 is NUMBER of buffers, not an ID!
glBindVertexArray (vao)

' ---------------------------------------------------------------------
' VBO: Raw data...
' ---------------------------------------------------------------------

' Create and bind as current VBO for currently-bound VAO...

glGenBuffers (1, Varptr vbo)		' NB. 1 is NUMBER of buffers, not an ID!
glBindBuffer (GL_ARRAY_BUFFER, vbo)

' Set buffer data for currently-bound VBO...

glBufferData (GL_ARRAY_BUFFER, SizeOf (vertex_data), vertex_data, GL_STATIC_DRAW)

glEnableVertexAttribArray (0)
glEnableVertexAttribArray (1)

' ---------------------------------------------------------------------
' Vertex attribute pointers operate on the currently-bound VAO...
' ---------------------------------------------------------------------

' Vertex attribute pointers below refer to those declared in vertex.glsl:

' 	layout (location = 0) in vec3 aPos;
' 	layout (location = 1) in vec3 aColor;

' Vertices data format description (aPos)...

glVertexAttribPointer (0, FLOATS_PER_ATTRIBUTE, GL_FLOAT, GL_FALSE, VA_Stride, Null)

' Colour data format description (aColor)...

glVertexAttribPointer (1, FLOATS_PER_ATTRIBUTE, GL_FLOAT, GL_FALSE, VA_Stride, Byte Ptr (VA_Offset))

glBindBuffer (GL_ARRAY_BUFFER, 0)
glBindVertexArray (0)

glDisableVertexAttribArray (1)
glDisableVertexAttribArray (0)

' WINDING (clockwise) AND CULLING

glFrontFace (GL_CW)
glEnable (GL_CULL_FACE)

' CAMERA

view_matrix			= SMat4F.LookAt (position, New SVec3F (0.0, 0.0, 0.0), New SVec3F (0.0, 1.0, 0.0))
projection_matrix	= SMat4F.Perspective (fov, _RetroGraphicsWidth, _RetroGraphicsHeight, near, far)

Local uni_cam:Int	= glGetUniformLocation (program, "camera")
glUniformMatrix4fv (uni_cam, 1, False, Varptr view_matrix.a)

Local uni_prj:Int	= glGetUniformLocation (program, "projection")
glUniformMatrix4fv (uni_prj, 1, False, Varptr projection_matrix.a)

' SET CLS COLOUR

glClearColor (0.1, 0.2, 0.75, 1.0)

' MAIN LOOP

While Not AppWindow.GetKey (GLFW_KEY_ESCAPE)

	' CLEAR SCREEN/DEPTH BUFFER
	
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

	glUseProgram (program)
	
	' Using vertex attributes 0 and 1 from vertex.glsl...
	
	glEnableVertexAttribArray (0)
	glEnableVertexAttribArray (1)

	glBindVertexArray (vao)

		' Draw using currently-bound VAO...
		
		glDrawArrays (GL_TRIANGLES, 0, vertex_count)

	glBindVertexArray (0)
	
	' Done using vertex attributes 0 and 1...

	glDisableVertexAttribArray (1)
	glDisableVertexAttribArray (0)

	AppWindow.SwapBuffers ()

	PollSystem ()
	
Wend

AppWindow.SetShouldClose (True)

glDeleteBuffers (1, Varptr vbo)
glDeleteVertexArrays (1, Varptr vao)

End

' WINDOW

Function Retro3D:TGLFWWindow (width:UInt, height:UInt, full_screen:Int = False, msaa_level:Int = 0)

	TGLFWWindow.Hint (GLFW_SAMPLES, msaa_level)
	TGLFWWindow.Hint (GLFW_CONTEXT_VERSION_MAJOR, 3)
	TGLFWWindow.Hint (GLFW_CONTEXT_VERSION_MINOR, 3)
	TGLFWWindow.Hint (GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE)

	?MacOS ' Ewww...
		TGLFWWindow.Hint (GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE)
	?

	Local monitor:TGLFWMonitor = Null

	If full_screen Then monitor = TGLFWMonitor.GetPrimaryMonitor ()

	AppWindow = New RetroGameWindow.Create (width, height, "Demo", monitor)

	If Not AppWindow
		Print "Failed to create GLFW window!"
		End
	EndIf

	AppWindow.MakeContextCurrent ()

	gladLoadGL (glfwGetProcAddress)
	
	' DEPTH TESTING
	
	glEnable (GL_DEPTH_TEST)
	glDepthFunc (GL_LESS)
	
	AppWindow.GetFrameBufferSize (_RetroGraphicsWidth, _RetroGraphicsHeight)

	Return AppWindow

EndFunction

Type RetroGameWindow Extends TGLFWWindow
	
	Method OnFrameBufferSize (width:Int, height:Int)
		glViewport (0, 0, width, height)
		AppWindow.GetFrameBufferSize (_RetroGraphicsWidth, _RetroGraphicsHeight)
	EndMethod
	
EndType

Const RETRO_RAD_DIVIDER:Float = 180.0 / Pi
Const RETRO_DEG_DIVIDER:Float = Pi / 180.0

Function Degrees:Float (radians:Float)
	Return radians * RETRO_RAD_DIVIDER
EndFunction

Function Radians:Float (degrees:Float)
	Return degrees * RETRO_DEG_DIVIDER
EndFunction
